import pytest
import tomli

from ini2toml.plugins.profile_independent_tasks import remove_empty_table_headers
from ini2toml.plugins.setuptools_pep621 import (
    Directive,
    SetuptoolsPEP621,
    _ConfusingPackagesConfig,
    activate,
)
from ini2toml.translator import Translator


@pytest.fixture
def plugin():
    return SetuptoolsPEP621()


@pytest.fixture
def translator(plugin):
    return Translator(plugins=[activate], ini_parser_opts=dict(delimiters=("=", ": ")))


@pytest.fixture
def parse(translator):
    return lambda text: translator.loads(text)


@pytest.fixture
def convert(translator):
    return lambda irepr: translator.dumps(irepr)


example_normalise_keys = """\
[metadata]
summary = Automatically translates .cfg/.ini files into TOML
author_email = example@example
license-file = LICENSE.txt
long_description_content_type = text/x-rst; charset=UTF-8
home_page = https://github.com/abravalheri/ini2toml/
classifier = Development Status :: 4 - Beta
platform = any
"""
expected_normalise_keys = """\
[metadata]
description = "Automatically translates .cfg/.ini files into TOML"
author-email = "example@example"
license-files = "LICENSE.txt"
long-description-content-type = "text/x-rst; charset=UTF-8"
url = "https://github.com/abravalheri/ini2toml/"
classifiers = "Development Status :: 4 - Beta"
platforms = "any"
"""


def test_normalise_keys(plugin, parse, convert):
    obj = parse(example_normalise_keys)
    with pytest.warns(DeprecationWarning):
        obj = plugin.normalise_keys(obj)
    assert convert(obj) == expected_normalise_keys


# ----


example_convert_directives = """\
[metadata]
version = attr: mymodule.myfunc
classifiers = file: CLASSIFIERS.txt
description = file: README.txt

[options]
entry-points = file: ENTRYPOINTS.txt
packages = find_namespace:
"""
expected_convert_directives = """\
[metadata]
version = {attr = "mymodule.myfunc"}
classifiers = {file = ["CLASSIFIERS.txt"]}
description = {file = ["README.txt"]}

[options]
entry-points = {file = ["ENTRYPOINTS.txt"]}
packages = {find_namespace = ""}
"""


def test_convert_directives(plugin, parse, convert):
    doc = parse(example_convert_directives)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.apply_value_processing(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc) == expected_convert_directives


# ----


example_apply_value_processing = """\
[metadata]
version = 5.3  # comment
classifiers =
    Development Status :: 4 - Beta
    Programming Language :: Python
keywords = python, module
[options]
zip-safe = False  # comment
package-dir =
    =src
install-requires =
    importlib-metadata; python_version<"3.8"
    configupdater>=3,<=4
cmdclass =
    customcmd = custom_build.CustomCmd
    other = custom_build.OtherCmd  # Some command
[options.entry-points]
# For example:
console-scripts =
    putup = pyscaffold.cli:run  # CLI exec
pyscaffold.cli =
    # comment
    fibonacci = ini2toml.skeleton:run # comment
    awesome = pyscaffoldext.awesome.extension:AwesomeExtension
"""
expected_apply_value_processing = """\
[metadata]
version = "5.3" # comment
classifiers = [
    "Development Status :: 4 - Beta",
    "Programming Language :: Python",
]
keywords = ["python", "module"]

[options]
zip-safe = false # comment
package-dir = {"" = "src"}
install-requires = [
    'importlib-metadata; python_version<"3.8"',
    "configupdater>=3,<=4",
]

[options.cmdclass]
customcmd = "custom_build.CustomCmd"
other = "custom_build.OtherCmd" # Some command

["project:entry-points"]
# For example:

["project:entry-points"."pyscaffold.cli"]
# comment
fibonacci = "ini2toml.skeleton:run" # comment
awesome = "pyscaffoldext.awesome.extension:AwesomeExtension"

["project:scripts"]
putup = "pyscaffold.cli:run" # CLI exec
"""


def test_move_entry_points_and_apply_value_processing(plugin, parse, convert):
    doc = parse(example_apply_value_processing)
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.move_and_split_entrypoints(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    text = convert(doc).strip()
    assert text == expected_apply_value_processing.strip()


# ----

example_split_subtables = """\
[options.packages.find]
where = src
[project:entry-points]
# For example:
[project:scripts]
django-admin = django.core.management:execute_from_command_line
"""

expected_split_subtables = """\
[tool.setuptools.packages.find]
where = "src"

[project.entry-points]
# For example:

[project.scripts]
django-admin = "django.core.management:execute_from_command_line"
"""


def test_split_subtables(plugin, parse, convert):
    doc = parse(example_split_subtables.strip())
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.split_subtables(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    converted = convert(doc).strip()
    assert remove_empty_table_headers(converted) == expected_split_subtables.strip()


# ----

example_entrypoints_and_split_subtables = """\
[options.packages.find]
where = src
[options.entry-points]
# For example:
console-scripts =
    django-admin = django.core.management:execute_from_command_line
gui-scripts =
    project = my.module:function [extra-dep]
pyscaffold.cli =
    config = pyscaffold.extensions.config:Config
    interactive = pyscaffold.extensions.interactive:Interactive
"""

expected_entrypoints_and_split_subtables = """\
[tool.setuptools.packages.find]
where = "src"

[project.entry-points]
# For example:

[project.entry-points."pyscaffold.cli"]
config = "pyscaffold.extensions.config:Config"
interactive = "pyscaffold.extensions.interactive:Interactive"

[project.scripts]
django-admin = "django.core.management:execute_from_command_line"

[project.gui-scripts]
project = "my.module:function [extra-dep]"
"""


def test_entrypoints_and_split_subtables(plugin, parse, convert):
    doc = parse(example_entrypoints_and_split_subtables.strip())
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.move_and_split_entrypoints(doc)
    doc = plugin.split_subtables(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    converted = convert(doc).strip()
    expected = expected_entrypoints_and_split_subtables.strip()
    assert remove_empty_table_headers(converted) == expected


# ----


example_handle_license_files = """\
[metadata]
license-files = LICENSE.txt
"""

expected_handle_license_files = """\
[metadata]

[options]
license-files = ["LICENSE.txt"]
"""


def test_handle_license_files(plugin, parse, convert):
    doc = parse(example_handle_license_files.strip())
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.handle_license(doc)
    doc = plugin.remove_metadata_not_in_pep621(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_handle_license_files.strip()


example_handle_multiple_license_files = """\
[metadata]
license-files = LICENSE.txt, NOTICE.txt
"""

expected_handle_multiple_license_files = """\
[metadata]

[options]
license-files = ["LICENSE.txt", "NOTICE.txt"]
"""


def test_handle_multiple_license_files(plugin, parse, convert):
    doc = parse(example_handle_multiple_license_files.strip())
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.handle_license(doc)
    doc = plugin.remove_metadata_not_in_pep621(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_handle_multiple_license_files.strip()


example_handle_license = """\
[metadata]
license = MPL-2.0
"""

expected_handle_license = """\
[metadata]
license = {text = "MPL-2.0"}
"""


def test_handle_license(plugin, parse, convert):
    doc = parse(example_handle_license.strip())
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.handle_license(doc)
    doc = plugin.remove_metadata_not_in_pep621(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_handle_license.strip()


example_handle_license_and_files = """\
[metadata]
license = MPL-2.0  # comment
license-files = LICENSE.txt
"""

expected_handle_license_and_files = """\
[metadata]
license = {text = "MPL-2.0"} # comment

[options]
license-files = ["LICENSE.txt"]
"""


def test_handle_license_and_files(plugin, parse, convert):
    doc = parse(example_handle_license_and_files.strip())
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.handle_license(doc)
    doc = plugin.remove_metadata_not_in_pep621(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_handle_license_and_files.strip()


# ----


example_handle_packages_find = """\
[options]
packages = find_namespace:
[options.packages.find]
where = src
exclude =
    tests
"""

expected_handle_packages_find = """\
[options]

["options.packages.find"]
where = ["src"]
exclude = ["tests"]
namespaces = true
"""


def test_handle_packages_find(plugin, parse, convert):
    doc = parse(example_handle_packages_find.strip())
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.handle_packages_find(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_handle_packages_find.strip()


# ----

example_fix_extras_require = """\
[options.extras-require]
pyarrow:python_version>'3.4' =
    pyarray>=1.0.0,<2.0dev; os_name=='posix'
    colorama; os_name=='nt'
"""

expected_fix_extras_require = """\
[build-system]
requires = ["setuptools>=61.2"]
build-backend = "setuptools.build_meta"

["project:optional-dependencies"]
pyarrow = [
    "pyarray>=1.0.0,<2.0dev; os_name=='posix' and python_version>'3.4'",
    "colorama; os_name=='nt' and python_version>'3.4'",
]
"""


def test_fix_extras_require(plugin, parse, convert):
    doc = plugin.template()
    doc.update(parse(example_fix_extras_require.strip()))
    doc = plugin.apply_value_processing(doc)
    doc = plugin.move_options_missing_in_pep621(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.fix_extras_require(doc)
    doc.pop("tool", None)
    doc.pop("options", None)
    doc.pop("metadata", None)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    text = convert(doc).strip()
    assert text == expected_fix_extras_require.strip()


# ----


example_move_setup_requires = """\
[options]
setup-requires =
    setuptools>=46.1.0
    setuptools_scm>=5
"""

expected_move_setup_requires = """\
[build-system]
requires = [
    "setuptools>=46.1.0",
    "setuptools_scm>=5",
]
build-backend = "setuptools.build_meta"
"""


def test_move_setup_requires(plugin, parse, convert):
    doc = plugin.template()
    doc.update(parse(example_move_setup_requires.strip()))
    doc = plugin.apply_value_processing(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    with pytest.warns(DeprecationWarning, match="'setup_requires' is deprecated"):
        doc = plugin.move_setup_requires(doc)
    doc.pop("tool", None)
    doc.pop("options", None)
    doc.pop("metadata", None)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    text = convert(doc).strip()
    assert text == expected_move_setup_requires.strip()


# ----


example_move_tests_require1 = """\
[options]
tests-require =
    pytest
    pytest-cov
"""

expected_move_tests_require1 = """\
[build-system]
requires = ["setuptools>=61.2"]
build-backend = "setuptools.build_meta"

["project:optional-dependencies"]
testing = [
    "pytest",
    "pytest-cov",
]
"""

example_move_tests_require2 = """\
[options]
tests-require =
    pytest
    pytest-cov

[options.extras-require]
testing =
    pytest>=6.2.5
"""

expected_move_tests_require2 = """\
[build-system]
requires = ["setuptools>=61.2"]
build-backend = "setuptools.build_meta"

["project:optional-dependencies"]
testing = [
    "pytest>=6.2.5",
    "pytest-cov",
]
"""


def test_move_tests_require(plugin, parse, convert):
    for example, expected in [
        (example_move_tests_require1, expected_move_tests_require1),
        (example_move_tests_require2, expected_move_tests_require2),
    ]:
        doc = plugin.template()
        doc.update(parse(example.strip()))
        doc = plugin.apply_value_processing(doc)
        doc = plugin.move_options_missing_in_pep621(doc)
        print(doc)
        print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
        with pytest.warns(DeprecationWarning, match="'tests_require' is deprecated"):
            doc = plugin.move_tests_require(doc)
        doc.pop("tool", None)
        doc.pop("options", None)
        doc.pop("metadata", None)
        print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
        print(doc)
        text = convert(doc).strip()
        assert text == expected.strip()


# ----


example_dynamic = """\
[metadata]
version = attr: django.__version__
classifiers = file: classifiers.txt
description = file: readme.txt
long-description = file: readme.txt, desc.txt
long-description-content-type: plain/text

[options]
entry-points = file: entry-points.txt
"""

expected_dynamic = """\
[metadata]

dynamic = [
    "readme",
    "version",
    "classifiers",
    "description",
    "entry-points",
    "scripts",
    "gui-scripts",
]

["options.dynamic"]
readme = {file = ["readme.txt", "desc.txt"], content-type = "plain/text"}
version = {attr = "django.__version__"}
classifiers = {file = ["classifiers.txt"]}
description = {file = ["readme.txt"]}
entry-points = {file = ["entry-points.txt"]}
"""  # noqa


def test_directives(plugin, parse, convert):
    doc = parse(example_dynamic.strip())
    doc = plugin.apply_value_processing(doc)
    doc = plugin.move_options_missing_in_pep621(doc)
    assert isinstance(doc["metadata"]["version"], Directive)
    assert doc["metadata"]["version"].kind == "attr"

    assert isinstance(doc["metadata"]["classifiers"], Directive)

    assert isinstance(doc["metadata"]["description"], Directive)
    assert doc["metadata"]["description"].kind == "file"

    assert isinstance(doc["metadata"]["entry-points"], Directive)


def test_handle_dynamic(plugin, parse, convert):
    doc = parse(example_dynamic.strip())
    doc = plugin.apply_value_processing(doc)
    doc = plugin.merge_and_rename_long_description_and_content_type(doc)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.handle_dynamic(doc)
    doc.pop("tool", None)
    doc.pop("options", None)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc).strip() == expected_dynamic.strip()


# ----

example_dynamic_dependencies = """\
[options]
install_requires = file: requirements.txt
[options.extras_require]
dev = file: dev-requirements.txt
"""

expected_dynamic_dependencies = """\
[project]
dynamic = ["version", "dependencies", "optional-dependencies"]

[tool.setuptools.dynamic]
dependencies = {file = ["requirements.txt"]}
optional-dependencies.dev = {file = ["dev-requirements.txt"]}
"""


def test_dynamic_dependecies(plugin, parse, convert):
    doc = parse(example_dynamic_dependencies)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.normalise_keys(doc)
    doc = plugin.pep621_transform(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    converted = tomli.loads(convert(doc).strip())
    converted.pop("build-system", None)
    converted["tool"]["setuptools"].pop("include-package-data", None)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(converted)
    assert converted == tomli.loads(expected_dynamic_dependencies)


# ----

expected_empty = """\
[build-system]
requires = ["setuptools>=61.2"]
build-backend = "setuptools.build_meta"

[project]
dynamic = ["version"]

[tool.setuptools]
include-package-data = false
"""


def test_empty(translator, plugin, parse, convert):
    doc = parse("")
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.normalise_keys(doc)
    doc = plugin.pep621_transform(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    converted = convert(doc).strip()
    assert remove_empty_table_headers(converted) == expected_empty.strip()

    # Same thing but with the higher level API:
    text = translator.translate("", profile_name="setup.cfg")
    assert text.strip() == expected_empty.strip()


# ----


example_data_files = """
[options]
data-files =
    a = b
"""

expected_data_files = """\
[build-system]
requires = ["setuptools>=61.2"]
build-backend = "setuptools.build_meta"

[project]
dynamic = ["version"]

[tool.setuptools]
data-files = {a = ["b"]}
include-package-data = false
"""


def test_data_files(translator):
    # Same thing but with the higher level API:
    with pytest.warns(DeprecationWarning, match="'data-files' is deprecated"):
        text = translator.translate(example_data_files, profile_name="setup.cfg")
        assert remove_empty_table_headers(text.strip()) == expected_data_files.strip()


# ----

# Example taken from https://github.com/jaraco/zipp
example_deps_with_line_continuation = """
[options.extras_require]
testing =
    # upstream
    pytest >= 6
    pytest-checkdocs >= 2.4
    pytest-flake8
    pytest-black >= 0.3.7; \
        # workaround for jaraco/skeleton#22
        python_implementation != "PyPy"
    pytest-cov
    pytest-mypy; \
        # workaround for jaraco/skeleton#22
        # comment have several lines
        python_implementation != "PyPy" and \
        # another comment
        os_name == "unix"
    pytest-enabler >= 1.0.1
"""

expected_deps_with_line_continuation = """
["options.extras_require"]
testing = [
    # upstream
    "pytest >= 6",
    "pytest-checkdocs >= 2.4"
    "pytest-flake8",
    "pytest-black >= 0.3.7; python_implementation != \"PyPy\"", # workaround for jaraco/skeleton#22
    "pytest-cov",
    "pytest-mypy; python_implementation != \"PyPy\" and os_name == \"unix\"",
    # workaround for jaraco/skeleton#22
    # comment have several lines
    # another comment
    pytest-enabler >= 1.0.1
"""  # noqa


def test_deps_with_line_continuation(plugin, parse, convert):
    doc = parse(example_convert_directives)
    print(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    doc = plugin.apply_value_processing(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc) == expected_convert_directives


# ----

# Example taken from https://github.com/aio-libs/aiohttp/
# In general, you should either list packages explicit or use find,
# but some configs change and `options.packages.find` is left behind
example_explicit_packages_with_find = """\
[options]
packages = aiohttp
[options.packages.find]
exclude =
  examples
"""

# In those cases, `ini2toml` should remove  `[options.packages.find]`
# And warn about it
expected_explicit_packages_with_find = """\
[tool.setuptools]
packages = ["aiohttp"]
"""


def test_explicit_packages_with_find(plugin, parse, convert):
    doc = parse(example_explicit_packages_with_find)
    print(doc)
    doc = plugin.apply_value_processing(doc)
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    with pytest.warns(_ConfusingPackagesConfig, match="[options.packages.find]"):
        doc = plugin.handle_packages_find(doc)
    doc.rename("options", ("tool", "setuptools"))
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(doc)
    assert convert(doc) == expected_explicit_packages_with_find



from inspect import cleandoc

from ini2toml.plugins.profile_independent_tasks import (
    normalise_newlines,
    remove_empty_table_headers,
)


def test_terminating_line():
    assert normalise_newlines("a") == "a\n"
    assert normalise_newlines("a\n") == "a\n"
    assert normalise_newlines("a\n\n") == "a\n"


def test_remove_empty_table_headers():
    text = """
    [tools]
    [tools.setuptools]

    [tools.setuptools.package]


    [tools.setuptools.package.find]
    where = "src"
    """
    expected = """
    [tools.setuptools.package.find]
    where = "src"
    """

    assert remove_empty_table_headers(cleandoc(text)) == cleandoc(expected)



import logging

import pytest

from ini2toml.drivers import configparser, configupdater
from ini2toml.plugins import toml_incompatibilities as plugin
from ini2toml.translator import Translator

EXAMPLE_FLAKE8 = """
[flake8]
# Some sane defaults for the code style checker flake8
# black compatibility
max_line_length = 88
# E203 and W503 have edge cases handled by black
extend_ignore = E203, W503
exclude =
    src/pyscaffold/contrib
    .tox
    build
    dist
    .eggs
    docs/conf.py
"""

EXAMPLE_DEVPI = """
[devpi:upload]
# Options for the devpi: PyPI server and packaging tool
# VCS export must be deactivated since we are using setuptools-scm
no_vcs = 1
formats = bdist_wheel
"""

EXAMPLES = {
    "flake8": (".flake8", "flake8", EXAMPLE_FLAKE8),
    "flake8-setup.cfg": ("setup.cfg", "flake8", EXAMPLE_FLAKE8),
    "devpi-setup.cfg": ("setup.cfg", "devpi:upload", EXAMPLE_DEVPI),
}


@pytest.mark.parametrize("example", EXAMPLES.keys())
@pytest.mark.parametrize("convert", (configparser.parse, configupdater.parse))
def test_log_warnings(example, convert, caplog):
    """ini2toml should display a warning via the logging system"""
    profile, section, content = EXAMPLES[example]
    translator = Translator(plugins=[plugin.activate], ini_loads_fn=convert)
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        translator.translate(content, profile)
    expected = plugin._warning_text(profile, repr(section))
    assert expected in caplog.text



import logging
import sys
from inspect import cleandoc
from unittest.mock import MagicMock

import pytest

from ini2toml import cli
from ini2toml.plugins import ErrorLoadingPlugin
from ini2toml.profile import Profile, ProfileAugmentation


def test_guess_profile(caplog):
    available = ["best_effort", "setup.cfg", "mypy.ini", "pyscaffold/default.cfg"]
    with caplog.at_level(logging.DEBUG):
        assert cli.guess_profile("best_effort", "setup.cfg", available) == "best_effort"
        assert caplog.text == ""
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        assert cli.guess_profile(None, "myproj/setup.cfg", available) == "setup.cfg"
        assert "Profile not explicitly set" in caplog.text
        assert "'setup.cfg' inferred" in caplog.text
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        profile = cli.guess_profile(None, "~/.config/pyscaffold/default.cfg", available)
        assert profile == "pyscaffold/default.cfg"
        assert "Profile not explicitly set" in caplog.text
        assert "'pyscaffold/default.cfg' inferred" in caplog.text
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        profile = cli.guess_profile(None, "myfile.ini", available)
        assert "Profile not explicitly set" in caplog.text
        assert "using 'best_effort'" in caplog.text


def _profile(name, help_text="help_text"):
    return Profile(name, help_text)


def _aug(name, active, help_text="help_text"):
    return ProfileAugmentation(lambda x: x, active, name, help_text)


def test_parse_args(tmp_path):
    file = tmp_path / "file.ini"
    file.touch()
    args = f"{file} -p setup.cfg -D hello world -E ehllo owrld".split()
    profiles = [_profile(n) for n in "setup.cfg default.cfg".split()]
    aug = [_aug(n, True) for n in "hello world other".split()]
    aug += [_aug(n, False) for n in "ehllo owrld".split()]
    params = cli.parse_args(args, profiles, aug)
    assert params.profile == "setup.cfg"
    assert params.active_augmentations == {
        "hello": False,
        "world": False,
        "ehllo": True,
        "owrld": True,
    }
    assert "other" not in params.active_augmentations


def test_critical_logging_sets_log_level_on_error(monkeypatch, caplog):
    spy = MagicMock()
    monkeypatch.setattr(sys, "argv", ["-vv"])
    monkeypatch.setattr(logging, "basicConfig", spy)
    with pytest.raises(ValueError):
        with cli.critical_logging():
            raise ValueError
    _args, kwargs = spy.call_args
    assert kwargs["level"] == logging.DEBUG


def test_critical_logging_does_nothing_if_no_argv(monkeypatch, caplog):
    spy = MagicMock()
    monkeypatch.setattr(sys, "argv", [])
    monkeypatch.setattr(logging, "basicConfig", spy)
    with pytest.raises(ValueError):
        with cli.critical_logging():
            raise ValueError
    assert spy.call_args is None


def test_exceptions2exit():
    with pytest.raises(SystemExit):
        with cli.exceptions2exit():
            raise ValueError


def test_early_plugin_error(monkeypatch):
    err = MagicMock(side_effect=ErrorLoadingPlugin("fake"))
    monkeypatch.setattr("ini2toml.translator.list_all_plugins", err)

    # Remove all attached handlers, so `logging.basicConfig` can work
    for logger in (logging.getLogger(), logging.getLogger(cli.__package__)):
        monkeypatch.setattr(logger, "handlers", [])

    assert logger.getEffectiveLevel() in {logging.NOTSET, logging.WARNING}

    monkeypatch.setattr("sys.argv", ["ini2toml", "-vv"])
    with pytest.raises(SystemExit):
        cli.run()

    assert logger.getEffectiveLevel() == logging.DEBUG


def test_help(capsys):
    with pytest.raises(SystemExit):
        cli.run(["--help"])
    out, _ = capsys.readouterr()
    assert len(out) > 0
    text = " ".join(x.strip() for x in out.splitlines())
    expected_profile_desc = """\
    conversion. Available profiles:
    - 'best_effort': guess option value conversion based on the string format.
    - '.coveragerc': convert settings to 'pyproject.toml' equivalent.
    - 'setup.cfg': convert settings to 'pyproject.toml' based on :pep:`621`.
    - '.isort.cfg': convert settings to 'pyproject.toml' equivalent.
    - 'mypy.ini': convert settings to 'pyproject.toml' equivalent.
    - 'pytest.ini': convert settings to 'pyproject.toml' ('ini_options' table).
    """
    expected = " ".join(x.strip() for x in expected_profile_desc.splitlines())
    print(out)
    assert expected in text


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="pyproject-fmt is only available on Python 3.7+",
)
def test_auto_formatting(tmp_path, capsys):
    setupcfg = """
    [metadata]
    version = 42
    name = myproj
    """
    normal_output = """
    requires = ["setuptools>=61.2"]
    """
    expected = """
    requires = [
      "setuptools>=61.2",
    ]
    """

    # Check if the underlying function works
    formatted = cli.apply_auto_formatting(cleandoc(expected))
    assert formatted.strip() == cleandoc(expected).strip()

    (tmp_path / "setup.cfg").write_text(cleandoc(setupcfg), encoding="utf-8")
    assert (tmp_path / "setup.cfg").exists()

    # Check the output when formatting in off
    cli.run([str(tmp_path / "setup.cfg")])
    out, _ = capsys.readouterr()
    assert cleandoc(normal_output) in out
    assert cleandoc(expected) not in out

    # Check the output when formatting in on
    cli.run(["-F", str(tmp_path / "setup.cfg")])
    out, _ = capsys.readouterr()
    assert cleandoc(normal_output) not in out
    assert cleandoc(expected) in out



from textwrap import dedent

import pytest
import tomli

from ini2toml.drivers import full_toml, lite_toml
from ini2toml.plugins.pytest import activate
from ini2toml.translator import FullTranslator

EXAMPLES = {
    "simple": {
        "example": """\
            [pytest]
            minversion = 6.0
            addopts = -ra -q --cov ini2toml
            testpaths = tests
            python_files = test_*.py check_*.py example_*.py
            required_plugins = pytest-django>=3.0.0,<4.0.0 pytest-html pytest-xdist>=1.0.0
            norecursedirs =
                dist
                build
                .tox
            filterwarnings=
                error
                ignore:Please use `dok_matrix` from the `scipy\\.sparse` namespace, the `scipy\\.sparse\\.dok` namespace is deprecated.:DeprecationWarning
            """,  # noqa
        "expected": """\
            [pytest]
            [pytest.ini_options]
            minversion = "6.0"
            addopts = "-ra -q --cov ini2toml"
            testpaths = ["tests"]
            python_files = ["test_*.py", "check_*.py", "example_*.py"]
            required_plugins = ["pytest-django>=3.0.0,<4.0.0", "pytest-html", "pytest-xdist>=1.0.0"]
            norecursedirs = [
                "dist", 
                "build", 
                ".tox", 
            ]
            filterwarnings = [
                "error",
                'ignore:Please use `dok_matrix` from the `scipy\\.sparse` namespace, the `scipy\\.sparse\\.dok` namespace is deprecated.:DeprecationWarning',
            ]
            """,  # noqa
    },
    "multiline-addopts": {
        "example": """\
            [pytest]
            addopts = -ra -q
                      # comment
                      --cov ini2toml
            """,
        "expected": '''\
            [pytest.ini_options]
            addopts = """
            -ra -q
            --cov ini2toml"""
            ''',
    },
}


@pytest.mark.parametrize("example_name", EXAMPLES.keys())
@pytest.mark.parametrize("convert", [full_toml.convert, lite_toml.convert])
def test_pytest(example_name, convert):
    example = EXAMPLES[example_name]["example"]
    expected = EXAMPLES[example_name]["expected"]
    translator = FullTranslator(plugins=[activate], toml_dumps_fn=convert)
    out = translator.translate(dedent(example), "pytest.ini").strip()
    expected = dedent(expected).strip()
    print("expected=\n" + expected + "\n***")
    print("out=\n" + out)
    try:
        assert expected in out
    except AssertionError:
        # At least the Python-equivalents when parsing should be the same
        assert tomli.loads(expected) == tomli.loads(out)



from textwrap import dedent

import pytest

from ini2toml.errors import UndefinedProfile
from ini2toml.plugins import profile_independent_tasks
from ini2toml.profile import Profile, ProfileAugmentation
from ini2toml.translator import FullTranslator, LiteTranslator, Translator


def test_simple_example():
    example = """\
    # comment

    [section1]
    option1 = value
    option2 = value # option comments are considered part of the value

    # comment
    [section2] # inline comment
    # comment
    option3 = value
    [section3]
    """
    # Obs: TOML always add a space before a new section
    expected = """\
    # comment

    [section1]
    option1 = "value"
    option2 = "value # option comments are considered part of the value"

    # comment

    [section2] # inline comment
    # comment
    option3 = "value"

    [section3]
    """
    translator = Translator(plugins=[])
    # ensure profile exists
    translator["simple"]
    out = translator.translate(dedent(example), "simple")
    print(out)
    assert out == dedent(expected)


def test_parser_opts():
    example = """\
    : comment

    [section1]
    option1 - value
    option2 - value : option comments are considered part of the value

    : comment
    [section2] : inline comment
    : comment
    option3 - value
    [section3]
    """
    # Obs: TOML always add a space before a new section
    expected = """\
    # comment

    [section1]
    option1 = "value"
    option2 = "value : option comments are considered part of the value"

    # comment

    [section2] # inline comment
    # comment
    option3 = "value"

    [section3]
    """

    parser_opts = {"comment_prefixes": (":",), "delimiters": ("-",)}
    translator = Translator(plugins=[], ini_parser_opts=parser_opts)
    # ensure profile exists
    translator["simple"]
    out = translator.translate(dedent(example), "simple")
    print(out)
    assert out == dedent(expected)


def test_undefined_profile():
    translator = Translator()
    with pytest.raises(UndefinedProfile):
        translator.translate("", "!!--UNDEFINED ??? PROFILE--!!")


simple_setupcfg = """\
[metadata]
summary = Automatically translates .cfg/.ini files into TOML
author_email = example@example
license-file = LICENSE.txt
long_description_content_type = text/x-rst; charset=UTF-8
home_page = https://github.com/abravalheri/ini2toml/
classifier = Development Status :: 4 - Beta
platform = any
"""


def test_reuse_object():
    """Make sure the same translator object can be reused multiple times"""
    profile = Profile("setup.cfg")
    augmentations = []
    for task in ("normalise_newlines", "remove_empty_table_headers"):
        fn = getattr(profile_independent_tasks, task)
        aug = ProfileAugmentation(
            profile_independent_tasks.post_process(fn), active_by_default=True
        )
        augmentations.append(aug)

    translator = Translator(
        profiles=[profile], plugins=(), profile_augmentations=augmentations
    )
    active_augmentations = {aug.name: True for aug in augmentations}
    for _ in range(5):
        out = translator.translate(simple_setupcfg, "setup.cfg", active_augmentations)
        assert len(out) > 0
    processors = [
        *profile.post_processors,
        *profile.intermediate_processors,
        *profile.pre_processors,
    ]
    deduplicated = {getattr(p, "__name__", ""): p for p in processors}
    # Make sure there is no duplication in the processors
    assert len(processors) == len(deduplicated)


def test_deduplicate_plugins():
    plugins = [
        profile_independent_tasks.activate,
        profile_independent_tasks.activate,
    ]
    translator = Translator(plugins=plugins)
    assert len(translator.plugins) == 1


minimal_example = """\
# comment

[section1]
option1 = value
option2 = value # option comments are considered part of the value
"""


def test_lite_translator():
    parser_opts = {"inline_comment_prefixes": ["#"]}
    translator = LiteTranslator(plugins=[], ini_parser_opts=parser_opts)
    # ensure profile exists
    translator["simple"]
    out = translator.translate(dedent(minimal_example), "simple")
    assert "# comment" not in out
    assert "part of the value" not in out


def test_full_translator():
    translator = FullTranslator(plugins=[])
    # ensure profile exists
    translator["simple"]
    out = translator.translate(dedent(minimal_example), "simple")
    assert "# comment" in out
    assert "part of the value" in out



import sys

if sys.version_info[:2] >= (3, 8):  # pragma: no cover
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version
else:  # pragma: no cover
    try:
        from importlib_metadata import PackageNotFoundError, version
    except ImportError:
        from pkg_resources import DistributionNotFound as PackageNotFoundError

        def version(name):
            import pkg_resources

            return pkg_resources.get_distribution(name).version


try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



import shutil
from pathlib import Path

TOC_TEMPLATE = """
Module Reference
================

The public API of ``ini2toml`` is mainly composed by ``ini2toml.api``.
Users may also find useful to import ``ini2toml.errors`` and
``ini2toml.types`` when handling exceptions or specifying type hints.

Plugin developers might try to use ``ini2toml.transformations``,
but they should take the stability of that module with a grain of salt...

.. toctree::
   :glob:
   :maxdepth: 2

   ini2toml.api
   ini2toml.errors
   ini2toml.types

.. toctree::
   :maxdepth: 1

   ini2toml.transformations

.. toctree::
   :caption: Plugins
   :glob:
   :maxdepth: 1

   plugins/*
"""

MODULE_TEMPLATE = """
``{name}``
~~{underline}~~

.. automodule:: {name}
   :members:{_members}
   :undoc-members:
   :show-inheritance:
"""


def gen_stubs(module_dir: str, output_dir: str):
    try_rmtree(output_dir)  # Always start fresh
    Path(output_dir, "plugins").mkdir(parents=True, exist_ok=True)
    for module in iter_public():
        text = module_template(module)
        Path(output_dir, f"{module}.rst").write_text(text, encoding="utf-8")
    for module in iter_plugins(module_dir):
        text = module_template(module, "activate")
        Path(output_dir, f"plugins/{module}.rst").write_text(text, encoding="utf-8")
    Path(output_dir, "modules.rst").write_text(TOC_TEMPLATE, encoding="utf-8")


def iter_public():
    lines = (x.strip() for x in TOC_TEMPLATE.splitlines())
    return (x for x in lines if x.startswith("ini2toml."))


def iter_plugins(module_dir: str):
    return (
        f'ini2toml.plugins.{path.with_suffix("").name}'
        for path in Path(module_dir, "plugins").iterdir()
        if path.is_file()
        and path.name not in {".", "..", "__init__.py"}
        and not path.name.startswith("_")
    )


def try_rmtree(target_dir: str):
    try:
        shutil.rmtree(target_dir)
    except FileNotFoundError:
        pass


def module_template(name: str, *members: str) -> str:
    underline = "~" * len(name)
    _members = (" " + ", ".join(members)) if members else ""
    return MODULE_TEMPLATE.format(name=name, underline=underline, _members=_members)
